/*

This is a program to print the status of an LSI 1030 RAID controller.
Copyright (C) 2004 CNET Networks, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA.

*/ 

#include <string.h>
#include <strings.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <pci/pci.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>

#include "mptctl.h"
#include "lsi/mpi_type.h"
#include "lsi/mpi.h"
#include "lsi/mpi_ioc.h"
#include "lsi/mpi_cnfg.h"

#define BIG 1024
#define REALLYBIG 10240

static int m = 0;

void *read_page(unsigned pagetype, unsigned pagewhich, unsigned pagenumber, int address, unsigned version) {
  char command[REALLYBIG];
  char reply[REALLYBIG];
  char sense[REALLYBIG];
  char *in;
  struct mpt_ioctl_command *cmd;
  Config_t *config_request;
  ConfigReply_t *config_reply;

  if(NULL == (in = calloc(REALLYBIG, 1))) {
    perror("calloc");
    exit(1);
  }

  cmd = (struct mpt_ioctl_command *) command;
  config_request = (Config_t *) cmd->MF;
  config_reply = (ConfigReply_t *) reply;

  bzero((void *) command, sizeof(command));
  bzero((void *) reply, sizeof(reply));
  bzero((void *) sense, sizeof(sense));
  bzero((void *) in, sizeof(in));

  cmd->hdr.iocnum = 0; /* XXX maybe you have more than one controller? */
  cmd->hdr.port = 0;
  cmd->hdr.maxDataSize =  BIG;
  cmd->timeout = 10;
  cmd->dataSgeOffset = (((char *) &config_request->PageBufferSGE) - ((char *) config_request)) / 4;
  cmd->senseDataPtr = (void *) sense;
  cmd->replyFrameBufPtr = (void *) config_reply;
  cmd->dataInBufPtr = (void *) in;
  cmd->maxSenseBytes = BIG;
  cmd->maxReplyBytes = BIG;
  cmd->dataInSize = BIG;
    
  config_request->Function = MPI_FUNCTION_CONFIG;
  config_request->Action = MPI_CONFIG_ACTION_PAGE_HEADER;
  config_request->Header.PageType = pagetype;
  config_request->Header.PageNumber = pagenumber;
  config_request->Header.PageLength = 0;
  config_request->Header.PageVersion = 0;
  config_request->PageAddress = address;

  if(-1 == ioctl(m, MPTCOMMAND, cmd)) {
    perror("MPI_FUNCTION_CONFIG (get page header)");
    exit(1);
  }

  if(version != config_reply->Header.PageVersion) {
    fprintf(stderr, "Version mismatch on pagetype %u, page number %u: expected %d, got %d\n",
	    pagetype,
	    pagenumber,
	    version,
	    config_reply->Header.PageVersion);
    exit(1);
  }

  config_request->Action = pagewhich;
  config_request->Header.PageLength = config_reply->Header.PageLength;

  if(config_reply->Header.PageLength > BIG) {
    fprintf(stderr, "reply too big\n");
    exit(1);
  }

  if(-1 == ioctl(m, MPTCOMMAND, cmd)) {
    perror("MPI_FUNCTION_CONFIG (get page)");
    exit(1);
  }

  return in;
}

int main() {
  RaidVolumePage0_t *page;
  RaidPhysDiskPage0_t *phys;
  char *VolumeTypes[] = { "IS", "IME", "IM" };
  int i;

  system("/sbin/modprobe mptctl");

  m = open("/dev/mptctl", O_RDWR);
  if(-1 == m) {
    int save_errno = errno;
    perror("open /dev/mptctl");
    if(save_errno == EACCES) {
      fprintf(stderr, "  You need to be root to run this program\n");
    } else if(save_errno == ENOENT) {
      fprintf(stderr, "  Try: mknod /dev/mptctl c 10 220\n");
    } else if(save_errno == ENODEV) {
      fprintf(stderr, "  Are you sure your controller is supported by mptlinux?\n");
    }
    exit(1);
  }

  /* XXX I don't know how it would work if you had multiple RAID volumes.
     Would those be separate pages?  And if so, how would I specify that I 
     wanted a different page? 

     If it's anything like physical disks, you would tweek `Address'. */
  page = read_page(MPI_CONFIG_PAGETYPE_RAID_VOLUME,
		   MPI_CONFIG_ACTION_PAGE_READ_CURRENT,
		   0,
		   0,
		   MPI_RAIDVOLPAGE0_PAGEVERSION);
  printf("ioc%d vol %d type %s, %d phy, %d GB", 
	 page->VolumeIOC, 
	 page->VolumeID,
	 page->VolumeType < sizeof(VolumeTypes) ? VolumeTypes[page->VolumeType] : "unknown",
	 page->NumPhysDisks,
	 page->MaxLBA / (2 * 1024 * 1024));
  if(page->VolumeStatus.Flags != 0) {
    printf(", flags");
    if(page->VolumeStatus.Flags & MPI_RAIDVOL0_STATUS_FLAG_ENABLED) printf(" ENABLED");
    if(page->VolumeStatus.Flags & MPI_RAIDVOL0_STATUS_FLAG_QUIESCED) printf(" QUIESCED");
    if(page->VolumeStatus.Flags & MPI_RAIDVOL0_STATUS_FLAG_RESYNC_IN_PROGRESS) printf(" RESYNC_IN_PROGRESS");
    if(page->VolumeStatus.Flags & MPI_RAIDVOL0_STATUS_FLAG_VOLUME_INACTIVE) printf(" VOLUME INACTIVE");
  }
  printf(", state");
  if(page->VolumeStatus.State == MPI_RAIDVOL0_STATUS_STATE_OPTIMAL) printf(" OPTIMAL");
  if(page->VolumeStatus.State & MPI_RAIDVOL0_STATUS_STATE_DEGRADED) printf(" DEGRADED");
  if(page->VolumeStatus.State & MPI_RAIDVOL0_STATUS_STATE_FAILED) printf(" FAILED");
  printf("\n");

  for(i = 0; i < page->NumPhysDisks; ++i) {
    phys = read_page(MPI_CONFIG_PAGETYPE_RAID_PHYSDISK,
		     MPI_CONFIG_ACTION_PAGE_READ_CURRENT,
		     0,
		     page->PhysDisk[i].PhysDiskNum,
		     MPI_RAIDPHYSDISKPAGE0_PAGEVERSION);
    char vendor[BIG];
    char productid[BIG];
    char rev[BIG];
    bzero(vendor, sizeof(vendor));
    bzero(productid, sizeof(productid));
    bzero(rev, sizeof(rev));
    strncpy(vendor,phys->InquiryData.VendorID,sizeof(phys->InquiryData.VendorID));
    strncpy(productid,phys->InquiryData.ProductID,sizeof(phys->InquiryData.ProductID));
    strncpy(rev,phys->InquiryData.ProductRevLevel,sizeof(phys->InquiryData.ProductRevLevel));
    
    printf("ioc%d phy %d %s %s %s, %d GB",
	   phys->PhysDiskIOC,
	   phys->PhysDiskNum,
	   vendor,
	   productid,
	   rev,
	   phys->MaxLBA / (2 * 1024 * 1024));
    if(phys->PhysDiskStatus.Flags) {
      printf(", flags");
      if(phys->PhysDiskStatus.Flags & MPI_PHYSDISK0_STATUS_FLAG_OUT_OF_SYNC) printf(" OUT_OF_SYNC");
      if(phys->PhysDiskStatus.Flags & MPI_PHYSDISK0_STATUS_FLAG_QUIESCED) printf(" QUIESCED");
    }
    printf(", state");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_ONLINE) printf(" ONLINE");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_MISSING) printf(" MISSING");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_NOT_COMPATIBLE) printf(" NOT COMPATIBLE");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_FAILED) printf(" FAILED");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_INITIALIZING) printf(" INITIALIZING");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_OFFLINE_REQUESTED) printf(" OFFLINE_REQUESTED");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_FAILED_REQUESTED) printf(" FAILED_REQUESTED");
    if(phys->PhysDiskStatus.State == MPI_PHYSDISK0_STATUS_OTHER_OFFLINE) printf(" OTHER_OFFLINE");
    printf("\n");
  }

  return 0;
}
